package de.tudresden.inf.st.mquat.solving.ilp;

import de.tudresden.inf.st.mquat.jastadd.model.ILP;
import de.tudresden.inf.st.mquat.jastadd.model.IlpVariable;
import de.tudresden.inf.st.mquat.solving.SolvingException;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class GurobiSolver extends ILPExternalSolver {

  /**
   * Create a new GLPK solver with default settings.
   * Default is:
   * <ul>
   *   <li>1 minute timeout</li>
   *   <li>delete temporary files on exit.</li>
   * </ul>
   * @see GurobiSolver#setDeleteFilesOnExit(boolean)
   */
  public GurobiSolver() {
  }


  @Override
  protected String[] getCommand(Path lp, Path solution, long remainingTimeForSolvingInMillis) {
    String[] command = {"gurobi_cl", " ResultFile=" + solution.toAbsolutePath(), "TimeLimit=" + remainingTimeForSolvingInMillis/1000, String.valueOf(lp.toAbsolutePath())};
    return command;
  }

  @Override
  protected void readFromPlainTextSolution(ILP ilp, Path solution, ILPSolution result,
                                                List<IlpVariable> variablesSetToOne) throws SolvingException {
    try (Stream<String> lines = Files.lines(solution)) {
      for (String line : lines.collect(Collectors.toList())) {
        if (line.startsWith("#")) {
          // comment line

          if (line.startsWith("# Objective Value =")) {
            result.setObjective(Double.valueOf(line.substring(20).trim()));
            logger.debug("read objective {}", Double.valueOf(line.substring(20).trim()));
          }
        } else {
          String[] tokens = line.split("\\s+");
          if (tokens.length == 2) {
            // tokens: name, value
            if (Math.round(Double.parseDouble(tokens[1])*1000000000)==1000000000) {
              logger.debug("found new variable {} with value {}", tokens[0], tokens[1]);
              IlpVariable variable = ilp.resolve(tokens[0]);
              if (variable == null) {
                throw new SolvingException("Could not find variable with name " + tokens[0]);
              }
              variablesSetToOne.add(variable);
            }
          }
        }
      }
    } catch (IOException e) {
      throw new SolvingException("Could not open solution file", e);
    } catch (NumberFormatException e) {
      throw new SolvingException("Could not parse solution file", e);
    }
  }

  @Override
  public String getName() {
    return "ilp-gurobi";
  }

}
